//
//  MAAnno.m
//  Medianno
//
//  Created by Matthias Neeracher on 8/10/11.
//  Copyright (c) 2011 Matthias Neeracher. All rights reserved.
//

#import "MAAnno.h"
#import "MATag.h"
#import "MATagDescription.h"
#import "MADocument.h"

#import <vector>

//
// Due to a bug in NSTokenField bindings, key value changes in tag descriptions don't
// propagate properly, because the token arrays are compared, rather than their display strings.
// our solution is to wrap the token arrays in a proxy class that will only compare equal by
// identity, never by structure.
//
@interface MANotAnArray : NSProxy {
    NSArray * contents;
}

- (id)initWithArray:(NSArray *)array;
+ (id)arrayWithArray:(NSArray *)array;
- (void)forwardInvocation:(NSInvocation *)invocation;

@end

@implementation MANotAnArray

- (id)initWithArray:(NSArray *)array
{
    contents = [array retain];

    return self;
}

- (void)dealloc
{
    [contents release];
    [super dealloc];
}

+ (id)arrayWithArray:(NSArray *)array
{
    return [[[MANotAnArray alloc] initWithArray:array] autorelease];
}
     
- (BOOL)isEqual:(id)object
{
    //
    // This is the only purpose of this class
    //
    return self==object;
}

- (NSMethodSignature *)methodSignatureForSelector:(SEL)sel
{
    return [contents methodSignatureForSelector:sel];
}

- (void)forwardInvocation:(NSInvocation *)invocation
{
    [invocation setTarget:contents];
    [invocation invoke];
}

@end

@implementation MAAnno
@dynamic location;
@dynamic notes;
@dynamic tags;
@dynamic media;

- (QTTime)qtLocation
{
    return QTTimeFromString(self.location);
}

- (NSArray *)tagDescriptions
{
    NSSet *                         tags   = self.tags;
    std::vector<NSManagedObject *> descs([tags count]);
    for (MATag * tag in tags) 
        descs[[tag seqNo]] = [tag tag];
    
    return [MANotAnArray arrayWithArray:[NSArray arrayWithObjects:&descs[0] count:descs.size()]];
 }

- (void)setTagDescriptions:(NSArray *)tagDescriptions
{
    NSManagedObjectContext *moc = [self managedObjectContext];
    
    NSSet * prevTags = self.tags;
    for (MATag * tag in prevTags)
        [moc deleteObject:tag];
    
    int seqNo = 0;
    std::vector<MATag *>  tags;
    for (MATagDescription * tagDesc in tagDescriptions) {
        MATag * tag     = [NSEntityDescription insertNewObjectForEntityForName:@"MATag" inManagedObjectContext:moc];
        tag.seqNo       = seqNo++;
        tag.tag         = tagDesc;
        tag.annotation  = self;
        tags.push_back(tag);
    }
    [self setTags:[NSSet setWithObjects:&tags[0] count:tags.size()]];
}

+ (NSSet *)keyPathsForValuesAffectingTagDescriptions
{
    return [NSSet setWithObject:@"tags"];
}

- (NSString *)exportText
{
    NSMutableString * text = [NSMutableString stringWithFormat:@"%@\t%@", self.location, self.notes];
    for (MATagDescription * tag in [self tagDescriptions])
        [text appendFormat:@"\t%@", [tag name]];
    return text;
}

- (MAAnno *)copyToDocument:(MADocument *)doc withMedia:(MAMedia *)media;
{
    MAAnno * anno               = [doc addAnnotationForMedia:media location:QTTimeFromString(self.location)];
    anno.notes                  = self.notes;
    NSMutableArray * tagDescs   = [NSMutableArray array];
    for (MATagDescription * tag in self.tagDescriptions)
        [tagDescs addObject:[tag copyToDocument:doc withMedia:media]];
    anno.tagDescriptions        = tagDescs;
    
    return anno;
}
@end
