//
//  MADocWindow.mm
//  Medianno
//
//  Created by Matthias Neeracher on 8/10/11.
//  Copyright 2011 Matthias Neeracher. All rights reserved.
//

#import <QTKit/QTKit.h>

#import "MADocWindow.h"
#import "MAAddMediaSheet.h"
#import "MADocument.h"
#import "MAMedia.h"
#import "MAAnno.h"
#import "MATag.h"
#import "MATagDescription.h"
#import "MAMovieWindow.h"
#import "MATagWindow.h"
#import "MAAppController.h"
#import "MADragging.h"

static const char * kMADocWindowObserver = "MADocWindowObserver";

@implementation MADocWindow

@synthesize exportAnnotations, exportTags, searchString;

- (id)initWithWindow:(NSWindow *)window
{
    self = [super initWithWindow:window];
    
    exportAnnotations   = YES;
    currentMovieTitle   = @"";
    
    return self;
}

- (void)dealloc
{
    [currentMovie release];
    [searchString release];
    [super dealloc];
}

- (void)windowDidLoad
{
    [super windowDidLoad];
    [mediaController setSortDescriptors:[NSArray arrayWithObject:[NSSortDescriptor sortDescriptorWithKey:@"date" ascending:YES]]];
    [annotationController setSortDescriptors:[NSArray arrayWithObject:[NSSortDescriptor sortDescriptorWithKey:@"location" ascending:YES]]];
    [NSTimer scheduledTimerWithTimeInterval:0.05 target:self selector:@selector(updateMovieTime:) userInfo:nil repeats:YES];
    [mediaController addObserver:self forKeyPath:@"selection.media" options:0 context:&kMADocWindowObserver];
    [mediaController addObserver:self forKeyPath:@"selection.name" options:0 context:&kMADocWindowObserver];
    [self registerOurDragTypes];
}

- (MAMovieWindow *)moviePanel
{
    return [[NSApp delegate] moviePanel];
}

- (MATagWindow *)tagPanel
{
    return [[NSApp delegate] tagPanel];    
}

- (void)moviePanelDidAppear
{
    [movieView setMovie:nil];
    if (![[self moviePanel] shouldDisplayMovie:currentMovie withTitle:currentMovieTitle])
        [movieView setMovie:currentMovie];
}

- (void)windowDidBecomeMain:(NSNotification *)notification
{
    [self moviePanelDidAppear];
    [[self tagPanel] setCurrentDocument:[self document]];
}

- (void)moviePanelDidClose
{
    [[self moviePanel] shouldDisplayMovie:nil withTitle:@""];
    [movieView setMovie:currentMovie];
}

- (void)windowDidResignMain:(NSNotification *)notification
{
    [self moviePanelDidClose];
    [[self tagPanel] setCurrentDocument:[[NSDocumentController sharedDocumentController] currentDocument]];
}

- (void)observeValueForKeyPath:(NSString *)keyPath ofObject:(id)object change:(NSDictionary *)change context:(void *)context
{
    if (context == &kMADocWindowObserver) {
        if (object == mediaController) {
            //
            // Current movie changed
            //
            [currentMovie release];
            if (MAMedia * currentMedia = [self currentMedia]) {
                NSURL * mediaURL    = [NSURL URLWithString:[currentMedia media]];
                currentMovie        = [[QTMovie alloc] initWithURL:mediaURL error:nil];
                currentMovieTitle   = [currentMedia name];
            } else {
                currentMovie        = nil;
                currentMovieTitle   = @"";
            }
            lastMovieTime           = QTMakeTime(-1, 600);
            [self updateMovieTime:nil];
        }
        [self moviePanelDidAppear];
    }
}

#pragma mark Media management

- (MAMedia *)currentMedia
{
    if ([[mediaController selectionIndexes] count] > 0) 
        return [[mediaController selectedObjects] objectAtIndex:0];
    else
        return nil;
}

- (QTTime)currentMovieTime
{
    return [currentMovie currentTime];
}

- (void)updateMovieTime:(NSTimer *)timer
{
    QTTime currentTime = [self currentMovieTime];
    if (QTTimeCompare(currentTime, lastMovieTime)) {
        [self willChangeValueForKey:@"currentMovieTime"];
        lastMovieTime = currentTime;
        [self didChangeValueForKey:@"currentMovieTime"];
    }
}

- (IBAction)addMediaFiles:(id)sender
{
    NSOpenPanel * openPanel = [NSOpenPanel openPanel];
    [openPanel setAllowedFileTypes:[QTMovie movieFileTypes:QTIncludeCommonTypes]];
    [openPanel setAllowsMultipleSelection:YES];
    [openPanel setCanChooseDirectories:YES];
    [openPanel beginSheetModalForWindow:[self window] completionHandler:^(NSInteger result) {
        [openPanel orderOut:self];
        if (result == NSFileHandlingPanelOKButton)
            [self addMedia:[openPanel URLs]];
    }];
}

- (void)addMediaTypes:(NSArray *)types inDirectory:(NSURL *)dir toList:(NSMutableArray *)urls
{
    NSFileManager *         fm      = [NSFileManager defaultManager];
    NSDirectoryEnumerator * dirEnum = [fm enumeratorAtURL:dir 
                               includingPropertiesForKeys:[NSArray arrayWithObjects:NSURLNameKey, NSURLIsDirectoryKey, nil]
                                                  options:NSDirectoryEnumerationSkipsHiddenFiles errorHandler:nil];
    for (NSURL * url in dirEnum) {
        NSNumber * isDirectory;            
        [url getResourceValue:&isDirectory forKey:NSURLIsDirectoryKey error:NULL];
        if (![isDirectory boolValue]) {
            NSString * ext = [url pathExtension];
            if ([types indexOfObjectPassingTest:^BOOL(id obj, NSUInteger idx, BOOL *stop) { 
                    return ![obj caseInsensitiveCompare:ext];
                }] != NSNotFound
            )
                [urls addObject:url];
        }
    }
}

- (void)addMedia:(NSArray *)urls
{
    NSMutableArray * expandedURLs = [NSMutableArray array];
    for (NSURL * url in urls) {
        if ([url isFileURL]) {
            NSNumber * isDirectory;            
            [url getResourceValue:&isDirectory forKey:NSURLIsDirectoryKey error:nil];
            if ([isDirectory boolValue]) {
                [self addMediaTypes:[QTMovie movieFileTypes:QTIncludeCommonTypes] inDirectory:url toList:expandedURLs];
                continue;
            }
        }
        [expandedURLs addObject:url];
    }
    [[[MAAddMediaSheet alloc] init] runWithParentWindow:self media:expandedURLs];
}

- (void)exportMediaToURL:(NSURL *)url
{
    NSIndexSet *    selection       = [annotationController selectionIndexes];
    NSArray *       annotations     = [annotationController arrangedObjects];
    //
    // Determine overall range of interest
    //
    QTTimeRange     exportRange     = QTMakeTimeRange(QTMakeTimeWithTimeInterval(0.0), [[[currentMovie movieAttributes] valueForKey:QTMovieDurationAttribute] QTTimeValue]);
    if ([selection count]) {
        if ([selection lastIndex] < [annotations count]-1)
            exportRange.duration    = [[annotations objectAtIndex:[selection lastIndex]+1] qtLocation];
        if ([selection firstIndex] > 0) {
            exportRange.time        = [[annotations objectAtIndex:[selection firstIndex]] qtLocation];
            exportRange.duration    = QTTimeDecrement(exportRange.duration, exportRange.time);
        }
    } 
    NSError *       error;
    QTMovie *       exportMovie     = [[QTMovie alloc] initWithMovie:currentMovie timeRange:exportRange error:&error];
    if (!exportMovie)
        [[self document] presentError:error];
    
    if ([selection count] < [selection lastIndex]-[selection firstIndex]+1) {
        [exportMovie setAttribute:[NSNumber numberWithBool:YES] forKey:QTMovieEditableAttribute];   
        //
        // Cut out discontiguous portions (in reverse order to simplify time accounting)
        //
        __block NSUInteger lastUsed = [selection lastIndex];
        [selection enumerateRangesWithOptions:NSEnumerationReverse usingBlock:^(NSRange range, BOOL *stop) {
            if (range.location+range.length < lastUsed+1) {
                QTTimeRange unusedRange = QTMakeTimeRange([[annotations objectAtIndex:range.location+range.length] qtLocation],
                                                          [[annotations objectAtIndex:lastUsed] qtLocation]);
                unusedRange.duration    = QTTimeDecrement(unusedRange.duration, unusedRange.time);
                unusedRange.time        = QTTimeDecrement(unusedRange.time, exportRange.time);
                [exportMovie deleteSegment:unusedRange]; 
            }
            lastUsed = range.location;
        }];
    }
    //
    // Now write out the movie
    //
    NSDictionary *  attributes      = [NSDictionary dictionaryWithObject:[NSNumber numberWithBool:YES] 
                                                                  forKey:QTMovieFlatten];
    [exportMovie writeToFile:[url path] withAttributes:attributes];
    [exportMovie release];
}

#define kMAAudioTypes   @"m4a", @"aiff", @"aif", @"wav", @"caf", @"mp3"
#define kMAVideoTypes   @"mov", @"mp4"

- (IBAction)exportMedia:(id)sender
{
    NSSavePanel * savePanel     = [NSSavePanel savePanel];
    NSDictionary *attrs         = [currentMovie movieAttributes];
    BOOL          hasVideo      = [[attrs objectForKey:QTMovieHasVideoAttribute] boolValue];
    BOOL          hasAudio      = !hasVideo || [[attrs objectForKey:QTMovieHasAudioAttribute] boolValue];
    [savePanel setAllowedFileTypes:(hasVideo 
                                    ? (hasAudio 
                                        ? [NSArray arrayWithObjects: kMAVideoTypes, kMAAudioTypes, nil]
                                        : [NSArray arrayWithObjects: kMAVideoTypes, nil])
                                    : [NSArray arrayWithObjects: kMAAudioTypes, nil])];
    [savePanel beginSheetModalForWindow:[self window] completionHandler:^(NSInteger result) {
        [savePanel orderOut:self];
        if (result == NSFileHandlingPanelOKButton) 
            [self exportMediaToURL:[savePanel URL]];
    }];
}

- (void)delete:(id)sender
{
    NSResponder * responder = [[self window] firstResponder];
    while (responder && responder != self) {
        if (responder == mediaTable) {
            [mediaController remove:sender];
            break;
        } else if (responder == annotationTable) {
            [annotationController remove:sender];
            break;
        }
        responder = [responder nextResponder];
    }
}

- (IBAction)toggleMediaPlay:(id)sender
{
    if ([currentMovie rate] > 0.0f)
        [currentMovie stop];
    else 
        [currentMovie play];
}

- (BOOL)validateUserInterfaceItem:(id <NSValidatedUserInterfaceItem>)item
{
    if ([[self tagPanel] validateUserInterfaceItem:item])
        return YES;
    
    if ([item action] == @selector(addMediaFiles:) 
     || [item action] == @selector(exportText:)
     || [item action] == @selector(importText:)
    )
        return YES;
  
    if ([item action] == @selector(paste:))
        return [self canPaste];
    
    if ([item action] == @selector(toggleMediaPlay:)) {
        NSString * title;
        if ([currentMovie rate] > 0.0f)
            title = NSLocalizedString(@"Pause", @"Pause playback");
        else 
            title = NSLocalizedString(@"Play",  @"Start playback");
        if ([(NSObject *)item isKindOfClass:[NSToolbarItem class]]) {
            NSToolbarItem * toolbarItem = (NSToolbarItem *)item;
            [toolbarItem setLabel:title];
            if ([currentMovie rate] > 0.0f)
                [toolbarItem setImage:[NSImage imageNamed:@"pause.icns"]];
            else 
                [toolbarItem setImage:[NSImage imageNamed:@"play.icns"]];
        } else {
            [(NSMenuItem *)item setTitle:title];
        }
    }
    return [mediaController selectionIndex] != NSNotFound;
}

#pragma mark Media navigation

static NSTimeInterval sLastSkip = 0.0;

- (void)resetSkipFactor:(id)sender
{
    sLastSkip = 0.0;
}

- (void)skipTimeInterval
{
    QTTime      interval        = QTMakeTimeWithTimeInterval(abs(sLastSkip));
    QTTime      current         = [currentMovie currentTime];
    if (sLastSkip > 0)
        current = QTTimeIncrement(current, interval);
    else
        current = QTTimeDecrement(current, interval);
    [currentMovie setCurrentTime:current];
    sLastSkip *= 1.1;
    [NSRunLoop cancelPreviousPerformRequestsWithTarget:self selector:@selector(resetSkipFactor:) object:self];
    [self performSelector:@selector(resetSkipFactor:) withObject:self afterDelay:0.5];
    [self updateMovieTime:nil];
}

- (IBAction)mediaSkipBackward:(id)sender
{
    if (sLastSkip >= 0)
        sLastSkip = -1.0;
    [self skipTimeInterval];
}

- (IBAction)mediaSkipForward:(id)sender
{
    if (sLastSkip <= 0)
        sLastSkip = 1.0;
    [self skipTimeInterval];
}

#pragma mark Tag token manipulation

- (NSString *)tokenFieldCell:(NSTokenFieldCell *)tokenFieldCell displayStringForRepresentedObject:(id)representedObject
{
    return [representedObject name];
}

- (NSArray *)tokenFieldCell:(NSTokenFieldCell *)tokenFieldCell completionsForSubstring:(NSString *)substring indexOfToken:(NSInteger)tokenIndex indexOfSelectedItem:(NSInteger *)selectedIndex
{
    return [[self document] tagNamesMatchingPrefix:substring];
}

- (id)tokenFieldCell:(NSTokenFieldCell *)tokenFieldCell representedObjectForEditingString:(NSString *)editingString
{
    return [[self document] tagDescriptionForName:editingString];
}

#pragma mark Split management

- (CGFloat)splitView:(NSSplitView *)splitView constrainMaxCoordinate:(CGFloat)proposedMax ofSubviewAt:(NSInteger)dividerIndex
{
    return splitView.frame.size.height-108.0;
}
 
- (CGFloat)splitView:(NSSplitView *)splitView constrainMinCoordinate:(CGFloat)proposedMin ofSubviewAt:(NSInteger)dividerIndex
{
    return 80.0;
}

- (BOOL)splitView:(NSSplitView *)splitView shouldAdjustSizeOfSubview:(NSView *)subview
{
    //
    // Only resize the bottom view if window is resized or zoomed
    //
    return subview != [[splitView subviews] objectAtIndex:0];
}

#pragma mark Annotations

- (void)tableViewSelectionDidChange:(NSNotification *)notification
{
    NSArray * selection = [annotationController selectedObjects];
    if ([selection count])
        if (MAAnno * firstSelectedAnno = [selection objectAtIndex:0]) 
            [currentMovie setCurrentTime:[firstSelectedAnno qtLocation]];
}

- (IBAction)addAnnotation:(id)sender
{
    [currentMovie stop];
    QTTime location = [self currentMovieTime];
    MAAnno * anno   = [[self document] addAnnotationForMedia:[self currentMedia] location:location];
    [annotationController setSelectedObjects:[NSArray arrayWithObject:anno]];
    [annotationTable editColumn:[annotationTable columnWithIdentifier:@"tags"] 
                            row:[annotationController selectionIndex]
                      withEvent:nil select:YES];
}

- (void)exportTextToURL:(NSURL *)url
{
    NSMutableString * text = [NSMutableString string];
    if (exportTags) {
        NSSet * tagSet;
        if (!exportAnnotations) {
            tagSet = [[self document] allTags];
        } else {
            NSMutableSet * tags = [NSMutableSet set];
            NSArray * annos;
            if ([[annotationController selectionIndexes] count] > 0) 
                annos = [annotationController selectedObjects];
            else
                annos = [annotationController arrangedObjects];
            for (MAAnno * anno in annos)
                for (MATag * tag in [anno tags])
                    [tags addObject:[tag tag]];
            tagSet = tags;
        } 
        NSArray * tagsToExport = [tagSet sortedArrayUsingDescriptors:[NSArray arrayWithObject:[NSSortDescriptor sortDescriptorWithKey:@"name" ascending:YES]]];
        for (MATagDescription * tag in tagsToExport)
            [text appendFormat:@"%@\n", [tag exportText]];
        if (exportAnnotations)
            [text appendString:@"\n"];
    }
    if (exportAnnotations) {
        NSArray * annosToExport;
        if ([[annotationController selectionIndexes] count] > 0) 
            annosToExport = [annotationController selectedObjects];
        else 
            annosToExport = [annotationController arrangedObjects];
        for (MAAnno * anno in annosToExport)
            [text appendFormat:@"%@\n", [anno exportText]];
    }
    [text writeToURL:url atomically:YES encoding:NSUTF8StringEncoding error:nil];
}

- (IBAction)exportText:(id)sender
{
    NSSavePanel * savePanel = [NSSavePanel savePanel];
    [savePanel setAllowedFileTypes:[NSArray arrayWithObject:@"txt"]];
    [savePanel setAccessoryView:textExportAccessoryView];
    exportAnnotations = [mediaController selectionIndex] != NSNotFound && (exportAnnotations || !exportTags);
    exportTags        = exportTags || !exportAnnotations;
    [savePanel beginSheetModalForWindow:[self window] completionHandler:^(NSInteger result) {
        [savePanel orderOut:self];
        if (result == NSFileHandlingPanelOKButton) 
            [self exportTextToURL:[savePanel URL]];            
    }];
}

- (void)importTextFromURL:(NSURL *)url
{
    NSString *  text    = [NSString stringWithContentsOfURL:url usedEncoding:nil error:nil];
    NSArray *   lines   = [text componentsSeparatedByString:@"\n"];
    if ([lines count] > 0) {
        MADocument *            doc         = [self document];
        MAMedia *               media       = [self currentMedia];
        //
        // Detect whether we start with a tag
        //
        BOOL                    nextIsTag   = NO;
        NSString *              firstLine   = [lines objectAtIndex:0];
        NSRegularExpression *   timeCode    = [NSRegularExpression regularExpressionWithPattern:@"^\\d:\\d\\d:\\d\\d:\\d\\d" options:0 error:nil];
        if ([timeCode numberOfMatchesInString:firstLine options:0 range:NSMakeRange(0, [firstLine length])] > 0)
            nextIsTag   = YES;
        for (NSString * line in lines) {
            NSArray *   components  = [line componentsSeparatedByString:@"\t"];
            NSUInteger  numComp     = [components count];
            if ([line isEqual:@""]) {
                nextIsTag = YES;            
            } else if (nextIsTag) {
                if (!media) 
                    break;
                QTTime      location= QTTimeFromString([components objectAtIndex:0]);
                MAAnno *    anno    = [doc addAnnotationForMedia:media location:location];
                [anno setNotes:[components objectAtIndex:1]];
                NSMutableArray * tagDescs = [NSMutableArray array];
                for (NSUInteger i = 2; i<numComp; ++i)
                    [tagDescs addObject:[doc tagDescriptionForName:[components objectAtIndex:i]]];
                [anno setTagDescriptions:tagDescs];
            } else {
                NSString *  name    = [components objectAtIndex:0];
                NSString *  notes   = numComp > 1 ? [components objectAtIndex:1] : nil;
                [doc tagDescriptionForName:name notes:notes];
            }
        }
    }
}

- (IBAction)importText:(id)sender
{
    NSOpenPanel * openPanel = [NSOpenPanel openPanel];
    [openPanel setAllowedFileTypes:[NSArray arrayWithObject:@"txt"]];
    [openPanel beginSheetModalForWindow:[self window] completionHandler:^(NSInteger result) {
        [openPanel orderOut:self];
        if (result == NSFileHandlingPanelOKButton) 
            [self importTextFromURL:[openPanel URL]];            
    }];
}

#pragma mark Searching

- (void)setSearchString:(NSString *)str
{
    [self willChangeValueForKey:@"searchString"];
    if (searchString != str) {
        [searchString release];
        searchString = [str retain];
    }
    if (!searchString || [searchString isEqual:@""]) {
        [mediaController setFilterPredicate:nil];
        [annotationController setFilterPredicate:nil];
    } else {
        NSArray * matchingTags = [[self document] tagDescriptionsMatching:searchString];
        [mediaController setFilterPredicate:
         [NSPredicate predicateWithFormat:@"SUBQUERY(annotations,$a,$a.notes CONTAINS[cd] %@ OR ANY $a.tagDescriptions IN %@).@count > 0", 
          searchString, matchingTags]];
        [annotationController setFilterPredicate:
         [NSPredicate predicateWithFormat:@"notes CONTAINS[cd] %@ OR ANY tagDescriptions IN %@",
          searchString, matchingTags]];
    }
    [self didChangeValueForKey:@"searchString"];
}
@end

@implementation MADocumentWindow

- (void)sendEvent:(NSEvent *)event
{
    
    if ([event type] == NSKeyDown && [[event characters] characterAtIndex:0]==3) 
        [[self windowController] addAnnotation:self];
    else
        [super sendEvent:event];
}

- (void)keyDown:(NSEvent *)event
{
    if ([[event characters] isEqualToString:@" "])
        [[self windowController] toggleMediaPlay:self];
    else
        [super keyDown:event];
}

@end
