//
//  MADragging.mm
//  Medianno
//
//  Created by Matthias Neeracher on 8/21/11.
//  Copyright 2011 Matthias Neeracher. All rights reserved.
//

#import "MADragging.h"
#import "MATagDescription.h"
#import "MAAnno.h"
#import "MAMedia.h"

#include <algorithm>

#define kMADragType @"org.aereperennius.medianno"

static NSPointerArray * sValidPasteboardData;

@interface MAPasteboardData : NSObject <NSPasteboardWriting> {
    NSArray * contents;
}

- (id)initWithArray:(NSArray *)array;
+ (id)pasteboardDataWithArray:(NSArray *)array;
+ (NSArray *)arrayFromPasteboard:(NSPasteboard *)pboard;

@end
 
@implementation MAPasteboardData

- (id)initWithArray:(NSArray *)array
{
    if (!sValidPasteboardData)
        sValidPasteboardData = [[NSPointerArray alloc] initWithOptions:NSPointerFunctionsZeroingWeakMemory];
    self        = [super init];
    contents    = array;
    if ([sValidPasteboardData count] > 10)
        [sValidPasteboardData compact];
    [sValidPasteboardData addPointer:contents];
    
    return self;
}

+ (id)pasteboardDataWithArray:(NSArray *)array
{
    return [[[MAPasteboardData alloc] initWithArray:array] autorelease];
}

+ (NSArray *)arrayFromPasteboard:(NSPasteboard *)pboard
{
    if (NSNumber * drag = [pboard propertyListForType:kMADragType]) {
        NSArray * possibleArray = (NSArray *)[drag longValue];
        for (NSUInteger count = [sValidPasteboardData count]; count--; )
            if ([sValidPasteboardData pointerAtIndex:count] == possibleArray)
                return possibleArray;
    }
    return nil;
}

- (NSArray *)writableTypesForPasteboard:(NSPasteboard *)pasteboard
{
    MAMedia * media;
    if (![contents count] || [(media = [contents objectAtIndex:0]) class] != [MAMedia class])
        return [NSArray arrayWithObjects:kMADragType, kUTTypeUTF8PlainText, nil];
    else if ([[NSURL URLWithString:[media media]] isFileURL])
        return [NSArray arrayWithObjects:kMADragType, kUTTypeFileURL, nil];
    else
        return [NSArray arrayWithObjects:kMADragType, kUTTypeURL, nil];
}

- (NSPasteboardWritingOptions)writingOptionsForType:(NSString *)type pasteboard:(NSPasteboard *)pasteboard
{
    return NSPasteboardWritingPromised;
}

- (id)pasteboardPropertyListForType:(NSString *)type
{
    if ([type isEqual:kMADragType]) {
        return [NSNumber numberWithLong:(long)contents];
    } else if ([type isEqual:(NSString *)kUTTypeUTF8PlainText]) {
        NSMutableString * stringBuffer = [NSMutableString string];
        for (id obj in contents)
            [stringBuffer appendFormat:@"%@\n", [obj exportText]];
        return stringBuffer;
    } else if ([type isEqual:(NSString *)kUTTypeURL] || [type isEqual:(NSString *)kUTTypeFileURL]) {
        return [contents count]>0 ? [[contents objectAtIndex:0] media] : nil;
    }
    return nil;
}

@end

@implementation MADocWindow (Dragging)

- (void)registerOurDragTypes
{
    [mediaTable setDraggingSourceOperationMask:NSDragOperationCopy forLocal:NO];
    [mediaTable registerForDraggedTypes:
     [NSArray arrayWithObjects:kMADragType, NSFilenamesPboardType, kUTTypeURL, nil]];
    
    [annotationTable setDraggingSourceOperationMask:NSDragOperationEvery forLocal:NO];
    [annotationTable setVerticalMotionCanBeginDrag:NO];
    [annotationTable registerForDraggedTypes:
     [NSArray arrayWithObjects:kMADragType, kUTTypeUTF8PlainText, nil]];    
}

- (BOOL)tableView:(NSTableView *)tableView writeRowsWithIndexes:(NSIndexSet *)rowIndexes toPasteboard:(NSPasteboard *)pboard
{
    NSArrayController * controller = tableView==annotationTable ? annotationController : mediaController;
    NSArray * objects = [[controller arrangedObjects] objectsAtIndexes:rowIndexes];
    [pboard clearContents];
    [pboard writeObjects:[NSArray arrayWithObject:[MAPasteboardData pasteboardDataWithArray:objects]]];
    
    return YES;
}

- (IBAction)copy:(id)sender
{
    NSPasteboard* pboard    = [NSPasteboard pasteboardWithName:NSGeneralPboard];
    NSResponder * responder = [[self window] firstResponder];
    while (responder && responder != self) {
        if (responder == mediaTable) {
            [self tableView:mediaTable writeRowsWithIndexes:[mediaController selectionIndexes] toPasteboard:pboard];
            break;
        } else if (responder == annotationTable) {
            [self tableView:annotationTable writeRowsWithIndexes:[annotationController selectionIndexes] toPasteboard:pboard];
            break;
        }
        responder = [responder nextResponder];
    }
}
 
- (IBAction)cut:(id)sender
{
    [self copy:sender];
    [self delete:sender];
}

- (Class)classOfDraggedObjects:(NSArray *)dragObjects
{
    return [dragObjects count] ? [[dragObjects objectAtIndex:0] class] : nil;
}

- (NSDragOperation)tableView:(NSTableView *)tableView validateDrop:(id<NSDraggingInfo>)info proposedRow:(NSInteger)row proposedDropOperation:(NSTableViewDropOperation)dropOperation
{
    //
    // We don't allow within-table dragging
    //
    if (tableView == [info draggingSource])
        return NSDragOperationNone;
    NSPasteboard *  pboard = [info draggingPasteboard];
    id              drag;
    if ((drag = [MAPasteboardData arrayFromPasteboard:pboard])) {
        //
        // We don't allow media drops in annotation table
        //
        Class dragClass = [self classOfDraggedObjects:drag];
        if (tableView == annotationTable) {
            if (dragClass == [MAMedia class] || (dragClass == [MAAnno class] && ![self currentMedia]))
                return NSDragOperationNone; 
            [tableView setDropRow:row dropOperation:NSTableViewDropAbove];
        } else {
            if (dragClass == [MAMedia class] || dragClass == [MATagDescription class]) {
                [tableView setDropRow:row dropOperation:NSTableViewDropAbove];
            } else {
                row = std::min<NSInteger>(row, [[mediaController arrangedObjects] count]-1);
                [tableView setDropRow:row dropOperation:NSTableViewDropOn];
            }
        }
        return NSDragOperationCopy;
    } else if (tableView == mediaTable) {
        if ((drag = [pboard propertyListForType:NSFilenamesPboardType])) {
            [tableView setDropRow:row dropOperation:NSTableViewDropAbove];
            return NSDragOperationLink;
        } else if ((drag = [pboard stringForType:(NSString *)kUTTypeURL])
         || (drag = [pboard stringForType:(NSString *)kUTTypeFileURL])
        ) {
            [tableView setDropRow:row dropOperation:NSTableViewDropAbove];
            return NSDragOperationLink;
        }
    }
    return NSDragOperationNone;
}

- (BOOL)tableView:(NSTableView *)tableView acceptDrop:(id<NSDraggingInfo>)info row:(NSInteger)row dropOperation:(NSTableViewDropOperation)dropOperation
{
    NSPasteboard *  pboard  = [info draggingPasteboard];
    MADocument *    doc     = [self document];
    MAMedia *       media   = nil;
    
    if (tableView == annotationTable) {
        media   =   [self currentMedia];
    } else if (dropOperation == NSTableViewDropOn) {
        NSArray * allMedia  = [mediaController arrangedObjects];
        if (row >= 0 && row < [allMedia count])
            media           = [allMedia objectAtIndex:row];
    }
    
    id drag;
    if ((drag = [MAPasteboardData arrayFromPasteboard:pboard])) {
        for (id obj in drag)
            [obj copyToDocument:doc withMedia:media];
    } else if ((drag = [pboard propertyListForType:NSFilenamesPboardType])) {
        NSMutableArray * urls = [NSMutableArray array];
        for (NSString * path in drag)
            [urls addObject:[NSURL fileURLWithPath:path]];
        [self addMedia:urls];
    } else if ((drag = [pboard stringForType:(NSString *)kUTTypeURL]) 
            || (drag = [pboard stringForType:(NSString *)kUTTypeFileURL])
    ) {
        [self addMedia:[NSArray arrayWithObject:[NSURL URLWithString:drag]]];
    }
    return YES;
}

- (IBAction)paste:(id)sender
{
    NSPasteboard*           pboard  = [NSPasteboard pasteboardWithName:NSGeneralPboard];
    MADocument *            doc     = [self document];
    NSManagedObjectContext *moc     = [doc managedObjectContext];  
    
    id drag;
    if ((drag = [MAPasteboardData arrayFromPasteboard:pboard])) {
        MAMedia * media   = [self currentMedia];
        for (id obj in drag)
            if ([obj managedObjectContext] != moc)
                [obj copyToDocument:doc withMedia:media];
    } else if ((drag = [pboard propertyListForType:NSFilenamesPboardType])) {
        NSMutableArray * urls = [NSMutableArray array];
        for (NSString * path in drag)
            [urls addObject:[NSURL fileURLWithPath:path]];
        [self addMedia:urls];
    } else if ((drag = [pboard stringForType:(NSString *)kUTTypeURL]) 
               || (drag = [pboard stringForType:(NSString *)kUTTypeFileURL])
               ) {
        [self addMedia:[NSArray arrayWithObject:[NSURL URLWithString:drag]]];
    }
}

- (BOOL)canPaste
{
    NSPasteboard*           pboard  = [NSPasteboard pasteboardWithName:NSGeneralPboard];
    id drag;
    if ((drag = [MAPasteboardData arrayFromPasteboard:pboard])) {
        MAMedia * media   = [self currentMedia];
         if (![drag count])
            return NO;
        id obj = [drag objectAtIndex:0];
        if ([obj managedObjectContext] == [[self document] managedObjectContext])
            return NO;
        else 
            return [obj class] != [MAAnno class] || media;
    } else if ((drag = [pboard propertyListForType:NSFilenamesPboardType])) {
        return YES;
    } else if ((drag = [pboard stringForType:(NSString *)kUTTypeURL]) 
               || (drag = [pboard stringForType:(NSString *)kUTTypeFileURL])
    ) {
        return YES;
    }    
    return NO;
}

@end

@implementation MATagWindow (Dragging)

- (void)registerOurDragTypes
{
    [tagTable setDraggingSourceOperationMask:NSDragOperationEvery forLocal:NO];
    [tagTable setVerticalMotionCanBeginDrag:NO];
    [tagTable registerForDraggedTypes:
     [NSArray arrayWithObjects:kMADragType, kUTTypeUTF8PlainText, nil]];
}

- (BOOL)tableView:(NSTableView *)tableView writeRowsWithIndexes:(NSIndexSet *)rowIndexes toPasteboard:(NSPasteboard *)pboard
{
    NSArray * objects = [[tagController arrangedObjects] objectsAtIndexes:rowIndexes];
    [pboard clearContents];
    [pboard writeObjects:[NSArray arrayWithObject:[MAPasteboardData pasteboardDataWithArray:objects]]];    
    return YES;
}

- (IBAction)copy:(id)sender
{
    [self tableView:nil writeRowsWithIndexes:[tagController selectionIndexes] toPasteboard:[NSPasteboard pasteboardWithName:NSGeneralPboard]];
}

- (IBAction)cut:(id)sender
{
    [self copy:sender];
    [self delete:sender];
}

@end
